//Project: RoverControl
//Homepage: www.HomoFaciens.de
//Author Norbert Heinz
//Version: 2.0
//Creation date: 21.11.2011
//Copyright: (c) 2011 by Norbert Heinz
//Microcontroller: Atmega8-16PU
//This program is free software you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation version 3 of the License.
//This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
//You should have received a copy of the GNU General Public License along with this program if not, see http://www.gnu.org/licenses/
//
//Program controls 1 H-brige (drive motor), one power transistors (light) and three servos (steering left/right, camera height/angle)
//Program based on PowerSwitch (author Christian Starkjohann) and usb-servo (author Ronald Schaten) - Thanks!
//Program uses V-USB,  published  under GNU GPL v2, GNU GPL v3 or proprietary (CommercialLicense.txt), Copyright: (c) 2005 by OBJECTIVE DEVELOPMENT Software GmbH


#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/eeprom.h>
#include <avr/wdt.h>

#include "usbdrv/usbdrv.h"
#include "usbdrv/oddebug.h"
#include <util/delay.h>

#define EEPROM_LOCATION (void *)37


static int powerOn = 0;
static int steeringAngle = 255;
static int headLight = 0;
static int timeCounter01 = 0;
static int cameraHeight = 255;
static int verticalAngle = 255;
static int irLEDcounter = 0;
static int moduleIRid = 0;
static int controlBits = 0;
static int IRcommandToSend = 0;



usbMsgLen_t usbFunctionSetup(uchar setupData[8]){
static uchar    replyBuf[2];

  usbMsgPtr = replyBuf;
  usbRequest_t *rq = (void *)setupData;   // cast to structured data for parsing

  switch(rq->bRequest){
    case 0:                                 //ECHO
      replyBuf[0] = rq->wValue.bytes[0];
      replyBuf[1] = rq->wValue.bytes[1];
      return 2;

    case 1:                                 //Set servo steering angle
      steeringAngle = setupData[2];
      return 0;

    case 2:                                 //Drive forward for powerOn * 20 ms
      PORTB = PORTB & 0b11100111;
      PORTB = PORTB | 0b00001000;
      powerOn = setupData[2];
      return 0;

    case 3:                                 //Drive backward for powerOn * 20 ms
      PORTB = PORTB & 0b11100111;
      PORTB = PORTB | 0b00010000;
      powerOn = setupData[2];
      return 0;

    case 4:                                 //Switch headlight
      headLight = setupData[2];
      return 0;

    case 5:                                 //Set camera height
      cameraHeight = setupData[2];
      return 0;

    case 6:                                 //Set vertical angle of camera
      verticalAngle = setupData[2];
      return 0;

    case 10:                                 //Initialize
      steeringAngle = 80;
      cameraHeight = 91;
      verticalAngle = 80;
      headLight = 0;
      powerOn = 0;
      return 0;

    case 11:                                 //Turn off servos and lights - sleep mode
      steeringAngle = 0;
      cameraHeight = 0;
      verticalAngle = 0;
      headLight = 0;
      powerOn = 0;
      return 0;

    case 12:                                 //Ask for infrared ID in range
      if(controlBits & (1 << 0)){
        replyBuf[0] = moduleIRid;
        replyBuf[1] = 1;
      }
      else{
        replyBuf[0] = moduleIRid;
        replyBuf[1] = 0;
      }
      controlBits |= 0b00000001;
      controlBits &= 0b11110001;
      irLEDcounter = 120;
      PORTC |= 0b00000100;
      IRcommandToSend = 0;
      moduleIRid = 0;
      return 2;

    case 13:                                 //Ask if new infrared ID is complete
      if(controlBits & (1 << 0)){
        replyBuf[0] = moduleIRid;
        replyBuf[1] = 1;
      }
      else{
        replyBuf[0] = moduleIRid;
        replyBuf[1] = 0;
      }
      return 2;

    case 14:                               //Send command via IR interface
      IRcommandToSend = setupData[2];
      controlBits |= 0b00000001;
      controlBits &= 0b11110001;
      irLEDcounter = 120;
      PORTC |= 0b00000100;
      moduleIRid = 0;
      return 0;

  }
  return 0;                               // ignore all unknown requests
}


//ISR(TIMER1_COMPA_vect) {
static void timerInterrupt(void){

  if(headLight > 0){
    PORTB = PORTB | 0b00100000;
  }

  timeCounter01++;
  if(timeCounter01 >= 5){
    timeCounter01 = 0;
    if(steeringAngle > 0 && steeringAngle < 180){
      PORTB |= 0b00000001;  //Turn servo pins on
    }
    if(verticalAngle > 0 && verticalAngle < 180){
      PORTD |= 0b00000001;  //Turn servo pins on
    }
    if(cameraHeight > 0 && cameraHeight < 180){
      PORTD |= 0b00000010;  //Turn servo pins on
    }
    if(powerOn > 0){
      powerOn--;
    }
  }

}

static void timerInterrupt_2(void){

  if(irLEDcounter>0){
    irLEDcounter--;
  }

}

int main(void){
uchar   i;

  //Port B, Pin 0 = Output, Servo control
  //Port B, Pin 3 = Output, H-Bridge driving motor forward
  //Port B, Pin 4 = Output, H-Bridge driving motor reverse
  //Port B, Pin 5 = Output, Power transistor light

  //Port B, Pin 6 = External oscillator
  //Port B, Pin 7 = External oscillator

  //Port C, Pin 1 = Input IR sensor
  //Port C, Pin 2 = Output IR diode

  //Port D, Pin 0 = Servo vertical angle camera
  //Port D, Pin 1 = Servo height angle camera
  //Port D, Pin 2 = Input, USB, Int0
  //Port D, Pin 6 = Input, USB D-
  //Port D, Pin 7 = Input, USB D+

  DDRB =  0b00111001;
  PORTB = 0b00000100;
  DDRC =  0b00000100;
  PORTC = 0b11111001;
  DDRD =  0b00000011;
  PORTD = 0b00111000;  //disable pullup resitors at usb-pins

  TCCR1A = 0x00;
  TCCR1B = (1<<CS12) | (1<<WGM12); //Configure Timer 0, prescaler 256
  OCR1A  = 188;                 //Interrupt after 12,000,000 / 256 / 188 = 249Hz (after 4.011ms)
  OCR1B = 188;                  //Interrupt after 12,000,000 / 256 / 188 = 249Hz (after 4.011ms)

  //Initialize timer 2 (8-bit) - IR communication
  OCR2 = 15;       //Compare value
  TCCR2 = ( 1 << WGM21 ) | ( 1 << CS22 );  //Switch to CTC mode, prescaler 64, 12 000 000/64/15 = 12500Hz = 0.08ms


  PORTB = PORTB | 0b00100000;

  usbDeviceDisconnect();  /* enforce re-enumeration, do this while interrupts are disabled! */
  i = 0;
  while(--i){             /* fake USB disconnect for > 500 ms */
    wdt_reset();
    _delay_ms(2);
  }
  usbDeviceConnect();
  usbInit();
  sei();

  PORTB = PORTB & 0b11011111;

  while(1 == 1){    /* main event loop */
    usbPoll();

    if(TCNT1 > steeringAngle){
      PORTB = PORTB & 0b11111110;  //Turn servo pin off
    }
    if(TCNT1 > verticalAngle){
      PORTD = PORTD & 0b11111110;  //Turn servo pin off
    }
    if(TCNT1 > cameraHeight){
      PORTD = PORTD & 0b11111101;  //Turn servo pin off
    }
    if(TCNT1 > headLight){
      PORTB = PORTB & 0b11011111;  //headlight off
    }

    if (TIFR & (1 << OCF1A)) {//'Manual' timer interrupt. Normal timer interrupt troubles USB-communication
      TIFR |= 1 << OCF1A;  /* clear pending flag */
      TCNT1 = 0;
      timerInterrupt();
    }

    if (TIFR & (1 << OCF2)) {//'Manual' timer interrupt. Normal timer interrupt troubles USB-communication
      TIFR |= 1 << OCF2;  /* clear pending flag */
      TCNT2 = 0;
      timerInterrupt_2();
    }

    if(powerOn == 0){//Turn off drive motor
      PORTB = PORTB & 0b11100111;
    }

    //Send and receive signals from IR interface
    if(controlBits & (1 << 0)){

      //Count response pulses
      if(!(PINC & (1 << 2)) && (controlBits & (1 << 1)) && !(controlBits & (1 << 3))){
        if((PINC & (1 << 1)) && (controlBits & (1 << 2))){
          controlBits &= 0b11111011;
          if(irLEDcounter < 50){
            irLEDcounter = 50;
          }
        }
        if(!(PINC & (1 << 1)) && !(controlBits & (1 << 2))){
          controlBits |= 0b00000100;
          moduleIRid++;
          irLEDcounter = 50;
        }
      }

      //Stop first IR pulse to initiate communication
      if(!(controlBits & (1 << 1)) && !(PINC & (1 << 2)) && (PINC & (1 << 1))){//Set controlbit 1 to start counting response pulses AFTER IR-LED was turned off AND Sensor is OFF
        controlBits |= 0b00000010;
        irLEDcounter = 200;//Longer pause to wait for response pulses
      }
      if(!(controlBits & (1 << 1)) && irLEDcounter < 6){//Turn off IR-LEDs
        PORTC &= 0b11111011;
      }

      if(controlBits & (1 << 3)){//Send IR command pulses
        if(irLEDcounter < 7 && IRcommandToSend > 0){
          if(PINC & (1 << 2)){
            PORTC &= 0b11111011;
            IRcommandToSend--;
            irLEDcounter = 30;
          }
          else{
            PORTC |= 0b00000100;
            irLEDcounter = 15;
          }
        }
      }

      if(irLEDcounter == 0){//Reset communication if nothing happened until irLEDcounter reached zero
        controlBits &= 0b11111001;
        if(IRcommandToSend > 0){//Initiate sending command pulses
          controlBits |= 0b00001000;
          irLEDcounter = 15;
        }
        else{
          controlBits &= 0b11110000;
          PORTC &= 0b11111011;
        }
      }
    }//if(controlBits & (1 << 0)){

  }
  return 0;
}

